import os.path as op
import re

from machotools.detect import is_macho
from machotools.macho_rewriter import rewriter_factory

LOADER_PATH = '@loader_path'


def build_rpath(root, lib_dir, file_path):
    """ Constructs an RPATH for a file.
    """
    parts = [LOADER_PATH]
    moves = [op.relpath(root, op.dirname(file_path)),
             op.relpath(lib_dir, root)]
    parts.extend([m for m in moves if m != '.'])
    return op.join(*parts)


def is_executable(path):
    """ Returns True if a file is a mach-o file
    """
    return is_macho(path)


def get_missing_libraries(path):
    """ Return a list of missing libraries for a given executable.
    """
    missing = []
    loader_path = op.dirname(path)
    rewriter = rewriter_factory(path)
    for dependency in rewriter.dependencies:
        if dependency.startswith(LOADER_PATH):
            dependency = loader_path + dependency[len(LOADER_PATH):]
        if not op.exists(dependency):
            missing.append(op.basename(dependency))
    return missing


def get_rpaths(path):
    """ Get all the RPATH entries for a mach-o file.
    """
    rewriter = rewriter_factory(path)
    rpaths = [p for p in rewriter.rpaths]
    return rpaths


def make_library_loads_relative(path, prefix):
    """ Rewrite all LC_LOAD_DYLIB commands in a given executable at `path` with
    names starting with `prefix` to @rpath<... everything after prefix>
    """
    if prefix[-1] == '/':
        prefix = prefix[:-1]
    prefix_len = len(prefix)

    with rewriter_factory(path) as rewriter:
        for dep in rewriter.dependencies:
            if not dep.startswith(prefix):
                continue
            replace = '@rpath' + dep[prefix_len:]
            rewriter.change_dependency(re.escape(dep), replace)

        # .dylib files should change their install_name too!
        if hasattr(rewriter, 'install_name'):
            install_name = rewriter.install_name
            if install_name.startswith(prefix):
                new_install_name = '@rpath' + install_name[prefix_len:]
                rewriter.install_name = new_install_name


def set_rpaths(path, rpaths):
    """ Set the RPATHs for a mach-o file.

    XXX: It looks like machotools always appends paths.
    """
    with rewriter_factory(path) as rewriter:
        rewriter.extend_rpaths(rpaths)



from __future__ import print_function

import argparse
import glob
import os
import os.path as op

from .api import is_executable, get_rpaths


def show_rpaths(path):
    """ Change the RPATH entry for a single executable.
    """
    shortpath = op.basename(path)
    if not is_executable(path):
        print('{} is not an executable file!'.format(shortpath))
        return

    rpaths = get_rpaths(path)
    if len(rpaths) > 0:
        print('{}: {}'.format(shortpath, ', '.join(rpaths)))


def walk_hierarchy(path, file_glob):
    """ Walk a directory hierarchy and call a function for each file whose name
    matches a given pattern.
    """
    for root, _, files in os.walk(path):
        for i, fn in enumerate(glob.fnmatch.filter(files, file_glob)):
            if i == 0:
                print('\nDirectory: {}'.format(root))
            show_rpaths(op.join(root, fn))


def main():
    description = 'Show the RPATH for all executables in a directory hierarchy'
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.HelpFormatter)
    parser.add_argument('--file-glob', '-g', default='*',
                        help='The glob used to locate executables.')
    parser.add_argument('--directory', '-d', required=True,
                        help='The root directory of the executables.')
    args = parser.parse_args()

    walk_hierarchy(op.abspath(args.directory), args.file_glob)


if __name__ == '__main__':
    main()



from __future__ import print_function

import argparse
from functools import partial
import glob
import os
import os.path as op

from .api import (build_rpath, is_executable, get_rpaths,
                  make_library_loads_relative, set_rpaths)


def visit_hierarchy(path, file_glob, visitor, **kwargs):
    """ Walk a directory hierarchy and call a function for each file whose name
    matches a given pattern.
    """
    for root, _, files in os.walk(path):
        for fn in glob.fnmatch.filter(files, file_glob):
            visitor(op.join(root, fn), **kwargs)


def fix_rpaths(root, lib_dirs, path, append=False, force=False):
    """ Change the RPATH entry for a single executable.
    """
    if not is_executable(path):
        print('{} is not an executable file!'.format(path))
        return

    for ld in lib_dirs:
        # NOTE: This is a no-op for ELF
        make_library_loads_relative(path, ld)

    rpaths = get_rpaths(path)
    if force or len(rpaths) > 0:
        new_rpaths = [build_rpath(root, ld, path) for ld in lib_dirs]
        if append:
            new_rpaths = rpaths + new_rpaths

        set_rpaths(path, new_rpaths)


def main():
    description = 'Fix the RPATH for all executables in a directory hierarchy'
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.HelpFormatter)
    parser.add_argument('--file-glob', '-g', default='*',
                        help='The glob used to locate executables.')
    parser.add_argument('--lib-dir', '-l', required=True, action='append',
                        help=('The directory where libraries are located. '
                              'This option can be supplied multiple times.'))
    parser.add_argument('--directory', '-d', required=True,
                        help='The root directory of the executables.')
    parser.add_argument('--append', '-a', action='store_true',
                        help='Append new RPATH to existing RPATH.')
    parser.add_argument('--force', '-f', action='store_true',
                        help=('Add an RPATH to files even if there previously '
                              'was not one.'))
    args = parser.parse_args()

    exe_root = op.abspath(op.normpath(args.directory))
    lib_dirs = [op.abspath(op.normpath(d)) for d in args.lib_dir]
    visitor = partial(fix_rpaths, exe_root, lib_dirs)
    visit_hierarchy(exe_root, args.file_glob, visitor,
                    append=args.append, force=args.force)


if __name__ == '__main__':
    main()



import os.path as op
import re
import subprocess
import sys

from elftools.common.exceptions import ELFError
from elftools.elf.dynamic import DynamicSection
from elftools.elf.elffile import ELFFile

PY3K = sys.version_info > (3, 0)
MISSING_LIB_REGEX = re.compile(r'.+(not found).+')


def _bytes2str(b):
    if PY3K:
        return b.decode('utf-8')
    else:
        return b


def _elf_type(path):
    with open(path, 'rb') as fp:
        elf = ELFFile(fp)
        return elf.header['e_type']


def _get_missing_libraries(path):
    """ Shell out to ldd to get the library references that can't be resolved.
    """
    cmdline = ["ldd", path]
    p = subprocess.Popen(cmdline, stdout=subprocess.PIPE)

    missing = []
    try:
        outs, _ = p.communicate()
        outs = _bytes2str(outs)
        for line in outs.split('\n'):
            if MISSING_LIB_REGEX.match(line):
                libname = line.split('=>')[0].strip()
                missing.append(libname)
    except subprocess.TimeoutExpired:
        p.kill()

    return missing


def build_rpath(root, lib_dir, file_path):
    """ Constructs an RPATH for a file.
    """
    parts = ['$ORIGIN']
    moves = [op.relpath(root, op.dirname(file_path)),
             op.relpath(lib_dir, root)]
    parts.extend([m for m in moves if m != '.'])
    return op.join(*parts)


def is_executable(path):
    """ Returns True if a file is an ELF file
    """
    try:
        with open(path, 'rb') as fp:
            ELFFile(fp)
        return True
    except ELFError:
        return False


def get_missing_libraries(path):
    """ Return a list of missing libraries for a given executable.
    """
    if not is_executable(path):
        return []

    return _get_missing_libraries(path)


def get_rpaths(path):
    """ Get all the RPATH and RUNPATH entries for an ELF binary
    """
    rpaths = []
    with open(path, 'rb') as fp:
        elf = ELFFile(fp)
        for section in elf.iter_sections():
            if isinstance(section, DynamicSection):
                for tag in section.iter_tags():
                    d_tag = tag.entry.d_tag
                    if d_tag == 'DT_RPATH':
                        rpaths.extend(tag.rpath.split(':'))
                    elif d_tag == 'DT_RUNPATH':
                        rpaths.extend(tag.runpath.split(':'))
    return rpaths


def make_library_loads_relative(path, with_prefix):
    """ Nothing to do. Just mirror the Mach-O function.
    """
    pass


def set_rpaths(path, rpaths):
    """ Set the RPATH for an ELF binary by shelling out to patchelf.
    """
    rpaths = ":".join(rpaths)
    cmdline = ["patchelf", "--set-rpath", rpaths, path]
    subprocess.check_call(cmdline)



from platform import system

__sys_os = system().lower()
if __sys_os == 'linux':
    from .elf import (build_rpath, is_executable, get_missing_libraries,
                      get_rpaths, make_library_loads_relative, set_rpaths)
elif __sys_os == 'darwin':
    from .macho import (build_rpath, is_executable, get_missing_libraries,
                        get_rpaths, make_library_loads_relative, set_rpaths)
else:
    raise ImportError('This platform({}) is not supported!'.format(__sys_os))


# Keep flake8 happy without telling it to ignore everything
hushup_flake8 = [build_rpath, is_executable, get_missing_libraries, get_rpaths,
                 make_library_loads_relative, set_rpaths]
del hushup_flake8



#! /usr/bin/env python

from platform import system
from setuptools import setup, find_packages

requirements = []
sys_os = system().lower()
if sys_os == 'linux':
    requirements.append('pyelftools')
elif sys_os == 'darwin':
    requirements.append('machotools')

setup(name='rpathology',
      version='0.0.5',
      license='MIT',
      author='John Wiggins',
      author_email='john.wiggins@xfel.eu',
      description='A library for dealing with executable RPATHs',
      long_description='',
      url='https://github.com/jwiggins/rpathology',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Operating System :: MacOS',
      ],
      packages=find_packages(),
      package_data={},
      entry_points={
          'console_scripts': [
              'rpath-fixer = rpathology.fixer:main',
              'rpath-missing = rpathology.list:main',
              'rpath-show = rpathology.show:main',
          ],
      },
      requires=requirements,
      )



from __future__ import print_function

import argparse
import glob
import os
import os.path as op

from .api import get_missing_libraries


def _show_misses(misses, root):
    print('Directory:', root)
    for path, missing in misses:
        libs = ', '.join(missing)
        print('{} is missing: {}'.format(op.basename(path), libs))
    print()


def walk_hierarchy(path, file_glob):
    """ Walk a directory hierarchy and check each file whose name matches a
    given pattern for missing library references.
    """
    for root, _, files in os.walk(path):
        misses = []
        for fn in glob.fnmatch.filter(files, file_glob):
            p = op.join(root, fn)
            miss = get_missing_libraries(p)
            if miss:
                misses.append((p, miss))

        if len(misses) > 0:
            _show_misses(misses, root)


def main():
    description = ('Show missing libraries for all executables in a directory '
                   'hierarchy')
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=argparse.HelpFormatter)
    parser.add_argument('--file-glob', '-g', default='*',
                        help='The glob used to locate executables.')
    parser.add_argument('--directory', '-d', required=True,
                        help='The root directory of the executables.')
    args = parser.parse_args()

    walk_hierarchy(op.abspath(args.directory), args.file_glob)


if __name__ == '__main__':
    main()
